@@ -30,7 +30,6 @@ module Client.State
   , clientScroll
   , clientDetailView
   , clientActivityBar
-  , clientShowMetadata
   , clientSubfocus
   , clientNextConnectionId
   , clientNetworkMap
@@ -40,6 +39,10 @@ module Client.State
   , clientExtensions
   , clientRegex
   , clientLogQueue
+  , clientActivityReturn
+  , clientErrorMsg
+  , clientLayout
+  , clientRtsStats
 
   -- * Client operations
   , withClientState
@@ -47,7 +50,9 @@ module Client.State
   , clientShutdown
   , clientPark
   , clientMatcher
+  , clientMatcher'
   , clientActiveRegex
+  , clientToggleHideMeta
 
   , consumeInput
   , currentCompletionList
@@ -82,6 +87,7 @@ module Client.State
   , retreatFocus
   , jumpToActivity
   , jumpFocus
+  , setExtraFocus
 
   -- * Scrolling
   , scrollClient
@@ -101,7 +107,6 @@ import           Client.Commands.WordCompletion
 import           Client.Configuration
 import           Client.Configuration.ServerSettings
 import           Client.Image.Message
-import           Client.Image.PackedImage
 import           Client.Image.Palette
 import           Client.Log
 import           Client.Message
@@ -130,6 +135,7 @@ import           Data.Map (Map)
 import qualified Data.Map as Map
 import           Data.Text (Text)
 import qualified Data.Text as Text
+import qualified Data.Text.Lazy as LText
 import           Data.Time
 import           Foreign.Ptr
 import           Foreign.StablePtr
@@ -139,6 +145,7 @@ import           Irc.Message
 import           Irc.RawIrcMsg
 import           Irc.UserInfo
 import           LensUtils
+import           RtsStats (Stats)
 import           Text.Regex.TDFA
 import           Text.Regex.TDFA.String (compile)
 
@@ -146,6 +153,7 @@ import           Text.Regex.TDFA.String (compile)
 data ClientState = ClientState
   { _clientWindows           :: !(Map Focus Window) -- ^ client message buffers
   , _clientPrevFocus         :: !Focus              -- ^ previously focused buffer
+  , _clientActivityReturn    :: !Focus              -- ^ focus prior to jumping to activity
   , _clientFocus             :: !Focus              -- ^ currently focused buffer
   , _clientSubfocus          :: !Subfocus           -- ^ current view mode
   , _clientExtraFocus        :: ![Focus]            -- ^ extra messages windows to view
@@ -165,8 +173,8 @@ data ClientState = ClientState
   , _clientScroll            :: !Int                      -- ^ buffer scroll lines
   , _clientDetailView        :: !Bool                     -- ^ use detailed rendering mode
   , _clientActivityBar       :: !Bool                     -- ^ visible activity bar
-  , _clientShowMetadata      :: !Bool                     -- ^ visible activity bar
   , _clientRegex             :: Maybe Regex               -- ^ optional persistent filter
+  , _clientLayout            :: !LayoutMode               -- ^ layout mode for split screen
 
   , _clientBell              :: !Bool                     -- ^ sound a bell next draw
 
@@ -174,6 +182,8 @@ data ClientState = ClientState
 
   , _clientExtensions        :: !ExtensionState           -- ^ state of loaded extensions
   , _clientLogQueue          :: ![LogLine]                -- ^ log lines ready to write
+  , _clientErrorMsg          :: Maybe Text                -- ^ transient error box text
+  , _clientRtsStats          :: Maybe Stats               -- ^ most recent GHC RTS stats
   }
 
 
@@ -239,19 +249,22 @@ withClientState cfg k =
         , _clientHeight            = 25
         , _clientEvents            = events
         , _clientPrevFocus         = Unfocused
+        , _clientActivityReturn    = Unfocused
         , _clientFocus             = Unfocused
         , _clientSubfocus          = FocusMessages
         , _clientExtraFocus        = []
         , _clientConfig            = cfg
         , _clientScroll            = 0
         , _clientDetailView        = False
-        , _clientShowMetadata      = True
         , _clientRegex             = Nothing
+        , _clientLayout            = view configLayout cfg
         , _clientActivityBar       = view configActivityBar cfg
         , _clientNextConnectionId  = 0
         , _clientBell              = False
         , _clientExtensions        = exts
         , _clientLogQueue          = []
+        , _clientErrorMsg          = Nothing
+        , _clientRtsStats          = Nothing
         }
 
 withExtensionState :: (ExtensionState -> IO a) -> IO a
@@ -295,7 +308,6 @@ recordChannelMessage network channel msg st
       , rendNicks       = HashSet.fromList (channelUserList network channel' st)
       , rendMyNicks     = highlights
       , rendPalette     = clientPalette st
-      , rendNickPadding = view (clientConfig . configNickPadding) st
       }
 
     -- on failure returns mempty/""
@@ -406,6 +418,7 @@ recordIrcMessage ::
   ClientMessage ->
   ClientState -> ClientState
 recordIrcMessage network target msg st =
+  updateTransientError msg $
   case target of
     TargetHidden      -> st
     TargetNetwork     -> recordNetworkMessage msg st
@@ -448,9 +461,28 @@ computeUserSigils network channel user =
          . csChannels . ix channel
          . chanUsers  . ix user
 
+
+-- | Detect /error/ messages and add the message text to the transient
+-- error display.
+updateTransientError :: ClientMessage -> ClientState -> ClientState
+updateTransientError msg =
+
+  let err = set clientErrorMsg . Just in
+
+  case view msgBody msg of
+    ErrorBody txt       -> err txt
+    IrcBody (Error txt) -> err txt
+    IrcBody (Reply code args)
+      | let info = replyCodeInfo code
+      , ErrorReply <- replyCodeType info ->
+          err (Text.intercalate " " (replyCodeText info : drop 1 args))
+    _ -> id
+
+
 -- | Record a message on a network window
 recordNetworkMessage :: ClientMessage -> ClientState -> ClientState
-recordNetworkMessage msg st = recordWindowLine focus wl st
+recordNetworkMessage msg st = updateTransientError msg
+                            $ recordWindowLine focus wl st
   where
     network    = view msgNetwork msg
     focus      | Text.null network = Unfocused
@@ -460,6 +492,7 @@ recordNetworkMessage msg st = recordWindowLine focus wl st
 
     cfg        = view clientConfig st
 
+
 -- | Record window line at the given focus creating the window if necessary
 recordWindowLine ::
   Focus ->
@@ -468,8 +501,9 @@ recordWindowLine ::
   ClientState
 recordWindowLine focus wl st = st2
   where
+    freshWindow = emptyWindow { _winHideMeta = view (clientConfig . configHideMeta) st }
     st1 = over (clientWindows . at focus)
-               (\w -> Just $! addToWindow wl (fromMaybe emptyWindow w))
+               (\w -> Just $! addToWindow wl (fromMaybe freshWindow w))
                st
 
     st2
@@ -480,27 +514,25 @@ recordWindowLine focus wl st = st2
 
       | otherwise = st1
 
-    hasMention = orOf (clientWindows . folded . winMention)
+    hasMention = elemOf (clientWindows . folded . winMention) WLImportant
 
 toWindowLine :: MessageRendererParams -> WindowLineImportance -> ClientMessage -> WindowLine
 toWindowLine params importance msg = WindowLine
   { _wlSummary    = msgSummary (view msgBody msg)
-  , _wlText       = msgText (view msgBody msg)
-  , _wlImage'     = _Image' # mkImage NormalRender
-  , _wlFullImage' = _Image' # mkImage DetailedRender
+  , _wlPrefix     = prefix
+  , _wlImage      = image
+  , _wlFullImage  = full
   , _wlImportance = importance
-  , _wlTimestamp  = zonedTimeToUTC (view msgTime msg)
+  , _wlTimestamp  = views msgTime packZonedTime msg
   }
   where
-    mkImage mode =
-      msgImage mode (view msgTime msg) params (view msgBody msg)
+    (prefix, image, full) = msgImage (view msgTime msg) params (view msgBody msg)
 
 -- | 'toWindowLine' but with mostly defaulted parameters.
 toWindowLine' :: Configuration -> WindowLineImportance -> ClientMessage -> WindowLine
 toWindowLine' config =
   toWindowLine defaultRenderParams
     { rendPalette     = view configPalette     config
-    , rendNickPadding = view configNickPadding config
     , rendMyNicks     = view configExtraHighlights config
     }
 
@@ -562,15 +594,22 @@ channelUserList network channel =
   views (clientConnection network . csChannels . ix channel . chanUsers) HashMap.keys
 
 
--- | Returns the current filtering predicate.
+-- | Returns the current filtering predicate if one is active.
 clientMatcher ::
-  ClientState {- ^ client state  -} ->
-  Text        {- ^ text to match -} ->
-  Bool        {- ^ is match      -}
+  ClientState          {- ^ client state       -} ->
+  Maybe (LText.Text -> Bool) {- ^ optional predicate -}
 clientMatcher st =
-  case clientActiveRegex st of
-    Nothing -> const True
-    Just r  -> matchTest r . Text.unpack
+  do r <- clientActiveRegex st
+     return (matchTest r . LText.unpack)
+
+-- | Strict version of 'clientMatcher'
+clientMatcher' ::
+  ClientState          {- ^ client state       -} ->
+  Maybe (Text -> Bool) {- ^ optional predicate -}
+clientMatcher' st =
+  do p <- clientMatcher st
+     return (p . LText.fromStrict)
+
 
 -- | Construct a text matching predicate used to filter the message window.
 clientActiveRegex :: ClientState -> Maybe Regex
@@ -612,11 +651,13 @@ Right urlPattern =
 
 -- | Find all the URL matches using 'urlPattern' in a given 'Text' suitable
 -- for being opened. Surrounding @<@ and @>@ are removed.
-urlMatches :: Text -> [Text]
+urlMatches :: LText.Text -> [Text]
 urlMatches txt = removeBrackets . extractText . (^?! ix 0)
-             <$> matchAll urlPattern (Text.unpack txt)
+             <$> matchAll urlPattern (LText.unpack txt)
   where
-    extractText (off,len) = Text.take len (Text.drop off txt)
+    extractText (off,len) = LText.toStrict
+                          $ LText.take (fromIntegral len)
+                          $ LText.drop (fromIntegral off) txt
 
     removeBrackets t =
       case Text.uncons t of
@@ -624,7 +665,7 @@ urlMatches txt = removeBrackets . extractText . (^?! ix 0)
        _                                  -> t
 
 -- | Remove a network connection and unlink it from the network map.
--- This operation assumes that the networkconnection exists and should
+-- This operation assumes that the network connection exists and should
 -- only be applied once per connection.
 removeNetwork :: NetworkId -> ClientState -> (NetworkState, ClientState)
 removeNetwork networkId st =
@@ -672,7 +713,7 @@ addConnection attempts lastTime network st =
 
 applyMessageToClientState ::
   ZonedTime                  {- ^ timestamp                -} ->
-  IrcMsg                     {- ^ message recieved         -} ->
+  IrcMsg                     {- ^ message received         -} ->
   NetworkId                  {- ^ message network          -} ->
   NetworkState               {- ^ network connection state -} ->
   ClientState                {- ^ client state             -} ->
@@ -779,14 +820,15 @@ clientExtraFocuses st =
 -- Some events like errors or chat messages mentioning keywords are
 -- considered important and will be jumped to first.
 jumpToActivity :: ClientState -> ClientState
-jumpToActivity st =
-  case mplus highPriority lowPriority of
-    Just (focus,_) -> changeFocus focus st
-    Nothing        -> st
+jumpToActivity st = changeFocus newFocus st
   where
     windowList   = views clientWindows Map.toAscList st
-    highPriority = find (view winMention . snd) windowList
+    highPriority = find (\x -> WLImportant == view winMention (snd x)) windowList
     lowPriority  = find (\x -> view winUnread (snd x) > 0) windowList
+    newFocus =
+      case mplus highPriority lowPriority of
+        Just (focus,_) -> focus
+        Nothing        -> view clientActivityReturn st
 
 -- | Jump the focus directly to a window based on its zero-based index.
 jumpFocus ::
@@ -799,6 +841,7 @@ jumpFocus i st
     windows   = view clientWindows st
     (focus,_) = Map.elemAt i windows
 
+
 -- | Change the window focus to the given value, reset the subfocus
 -- to message view, reset the scroll, remember the previous focus
 -- if it changed.
@@ -808,16 +851,45 @@ changeFocus ::
   ClientState
 changeFocus focus st
   = set clientScroll 0
+  . activateCurrent
+  . deactivatePrevious
   . updatePrevious
   . set clientFocus focus
   . set clientSubfocus FocusMessages
   $ st
   where
     oldFocus = view clientFocus st
+
     updatePrevious
       | focus == oldFocus = id
       | otherwise         = set clientPrevFocus oldFocus
 
+    -- always activate the new window. If it was already active this
+    -- will clear the marker.
+    activateCurrent = over (clientWindows . ix focus) windowActivate
+
+    -- Don't deactivate a window if it's going to stay active
+    deactivatePrevious
+      | oldFocus `elem` focus : view clientExtraFocus st = id
+      | otherwise = over (clientWindows . ix oldFocus) windowDeactivate
+
+
+-- | Unified logic for assigning to the extra focuses field that activates
+-- and deactivates windows as needed.
+setExtraFocus :: [Focus] -> ClientState -> ClientState
+setExtraFocus newFocuses st
+  = aux windowDeactivate newlyInactive
+  $ aux windowActivate   newlyActive
+  $ set clientExtraFocus newFocuses st
+  where
+    newlyActive = newFocuses \\ (view clientFocus st : view clientExtraFocus st)
+
+    newlyInactive = view clientExtraFocus st \\ (view clientFocus st : newFocuses)
+
+    aux f xs st1 =
+      foldl' (\acc w -> overStrict (clientWindows . ix w) f acc) st1 xs
+
+
 -- | Change the subfocus to the given value, preserve the focus, reset
 -- the scroll.
 changeSubfocus ::
@@ -825,7 +897,8 @@ changeSubfocus ::
   ClientState {- ^ client state -} ->
   ClientState
 changeSubfocus focus
-  = set clientScroll 0
+  = set clientErrorMsg Nothing
+  . set clientScroll 0
   . set clientSubfocus focus
 
 -- | Return to previously focused window.
@@ -912,3 +985,8 @@ clientAutoconnects st =
   [ network | (network, cfg) <- views (clientConfig . configServers) HashMap.toList st
             , view ssAutoconnect cfg
             ]
+
+-- | Toggle the /hide metadata/ setting for the focused window.
+clientToggleHideMeta :: ClientState -> ClientState
+clientToggleHideMeta st =
+  overStrict (clientWindows . ix (view clientFocus st) . winHideMeta) not st
